# ~ coding: future_fstrings ~

from __future__ import absolute_import, print_function, unicode_literals

# http://python-future.org/str_object.html <- habilita el tipo de str de py3
from future.builtins import super, zip, isinstance, bytes
from future.builtins import str as future_str
from future.utils import iteritems, with_metaclass

import decimal
import base64
import datetime
import abc
import sys
import os

from .exceptions import ValidationError


class _Missing(object):
    def __bool__(self):
        return False

    def __repr__(self):
        return "<m9g.missing>"


def _default_validation_hook(value):
    pass


missing = _Missing()


class Field(with_metaclass(abc.ABCMeta)):
    VALIDATE = os.environ.get('M9G_VALIDATE_TYPES', False)

    def __init__(
            self,
            default=missing,
            pk=False,
            allow_none=False,
            validation_hook=_default_validation_hook
    ):
        self.__pk = pk
        self.allow_none = allow_none
        self.validation_hook = validation_hook
        if default is not missing:
            default = self.adapt(default)
            self.validate(default)
        self.default = default

    @classmethod
    def _field_format_to_model_format(cls, format):
        if format == "json":
            return "jsondict"
        return format

    def serialize(self, format, value):
        serialize_fn = getattr(self, f"serialize_{format}", None)
        if not serialize_fn:
            return self.serialize_default(format, value)
        return serialize_fn(value)

    def deserialize(self, format, value):
        deserialize_fn = getattr(self, f"deserialize_{format}", None)
        if not deserialize_fn:
            return self.deserialize_default(format, value)
        return deserialize_fn(value)

    def serialize_default(self, format, value):
        return value

    def deserialize_default(self, format, value):
        return value

    def get_default(self):
        if callable(self.default):
            return self.default()
        return self.default

    @classmethod
    def is_field_type_of(cls, value):
        return isinstance(value, cls.FIELD_TYPE)

    def validate(self, value):
        if self.VALIDATE:
            self._validate(value)

    def _validate(self, value):
        if not self.allow_none and value is None:
            raise ValidationError(f"{self.__class__} was marked as not nullable")

        if value is not None and not isinstance(value, self.FIELD_TYPE):
            raise ValidationError(f"{value} is not instance of declared type {self.FIELD_TYPE}")

        validation_hook = self.validation_hook
        validation_hook(value)

    @property
    def pk(self):
        return self.__pk

    @pk.setter
    def pk(self, pk):
        self.__pk = pk

    def adapt(self, value):
        return value


class AtomicField(Field):
    def serialize_json(self, value):
        return value

    def deserialize_json(self, value):
        return value


class StringField(AtomicField):
    FIELD_TYPE = (str, future_str)

    def __init__(self, encoding='utf-8', **kwargs):
        self.encoding = encoding
        super().__init__(**kwargs)

    def adapt(self, value):
        if sys.version_info[0] < 3 and isinstance(value, str):
            return value.decode(self.encoding)
        return value


class IntField(AtomicField):
    FIELD_TYPE = int


class FloatField(AtomicField):
    FIELD_TYPE = float

    def adapt(self, value):
        if isinstance(value, int):
            return float(value)
        return value


class DecimalField(AtomicField):
    FIELD_TYPE = decimal.Decimal

    def __init__(self, quantize=decimal.Decimal('0.01'), **kwargs):
        super().__init__(**kwargs)
        self.quantize = quantize

    def adapt(self, value):
        if isinstance(value, (float, int)):
            return decimal.Decimal(value)
        return value

    def serialize_pickle(self, value):
        # usa deserialización default
        return value.quantize(self.quantize)

    def serialize_pydict(self, value):
        # usa deserialización default
        return value.quantize(self.quantize)

    def serialize_json(self, value):
        return str(value.quantize(self.quantize))

    def deserialize_json(self, value):
        return decimal.Decimal(value)


class BytesField(AtomicField):
    FIELD_TYPE = bytes

    def serialize_json(self, value):
        return base64.b64encode(value).decode('utf-8')

    def deserialize_json(self, value):
        return base64.b64decode(value.encode('utf-8'))


class DateField(AtomicField):
    FIELD_TYPE = datetime.date

    def serialize_json(self, value):
        return value.isoformat()

    def deserialize_json(self, date_str):
        date = datetime.datetime.strptime(date_str, "%Y-%m-%d")
        return date.date()

    @classmethod
    def is_field_type_of(cls, value):
        return type(value) == cls.FIELD_TYPE


class DateTimeField(AtomicField):
    FIELD_TYPE = datetime.datetime

    def serialize_json(self, value):
        # No usamos isoformat porque no agrega los milisegundos
        return value.strftime('%Y-%m-%dT%H:%M:%S.%f')

    def deserialize_json(self, value):
        return datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')

    @classmethod
    def is_field_type_of(cls, value):
        return type(value) == cls.FIELD_TYPE


class DictField(Field):
    FIELD_TYPE = dict

    def __init__(self, key_field_type, value_field_type, **kwargs):
        self.key_field_type = key_field_type
        self.value_field_type = value_field_type
        super().__init__(**kwargs)

    def get_default(self):
        if self.default == {}:
            return {}  # Return new dictionary
        return super().get_default()

    def serialize_default(self, format, to_serialize_dict):
        if isinstance(self.key_field_type, StringField) or format != "json":
            return self._serialize_default_as_dict(format, to_serialize_dict)
        else:
            return self._serialize_default_as_pair_list(format, to_serialize_dict)

    def deserialize_default(self, format, value):
        if isinstance(self.key_field_type, StringField) or format != "json":
            return self._deserialize_default_as_dict(format, value)
        else:
            return self._deserialize_default_as_pair_list(format, value)

    def _serialize_default_as_dict(self, format, to_serialize_dict):
        return dict(
            (self.key_field_type.serialize(format, k), self.value_field_type.serialize(format, v))
            for k, v in iteritems(to_serialize_dict)
        )

    def _deserialize_default_as_dict(self, format, str_dict):
        from_serialize = {}
        for k, v in iteritems(str_dict):
            deserialized_k = self.key_field_type.deserialize(format, k)
            deserialized_v = self.value_field_type.deserialize(format, v)
            from_serialize[deserialized_k] = deserialized_v

        return from_serialize

    def _serialize_default_as_pair_list(self, format, to_serialize_dict):
        serialized = []
        for k, v in iteritems(to_serialize_dict):
            serialized_k = self.key_field_type.serialize(format, k)
            serialized_v = self.value_field_type.serialize(format, v)
            serialized.append((serialized_k, serialized_v))

        return serialized

    def _deserialize_default_as_pair_list(self, format, pair_list):
        from_serialize = {}
        for k, v in pair_list:
            deserialized_k = self.key_field_type.deserialize(format, k)
            deserialized_v = self.value_field_type.deserialize(format, v)
            from_serialize[deserialized_k] = deserialized_v

        return from_serialize

    def _validate(self, value):
        super()._validate(value)
        if value is None:
            return  # allow_none validated in base class
        for k, v in iteritems(value):
            self.key_field_type.validate(k)
            self.value_field_type.validate(v)


class ListField(Field):
    FIELD_TYPE = list

    def __init__(self, items_field_type, **kwargs):
        self.items_fields_type = items_field_type
        super().__init__(**kwargs)

    def get_default(self):
        if self.default == []:
            return []
        return super().get_default()

    def serialize_default(self, format, value):
        return [self.items_fields_type.serialize(format, x) for x in value]

    def deserialize_default(self, format, value):
        return [self.items_fields_type.deserialize(format, x) for x in value]

    def _validate(self, value):
        super()._validate(value)
        if value is None:
            return  # allow_none validated in base class
        for v in value:
            self.items_fields_type.validate(v)


class TupleField(Field):
    FIELD_TYPE = tuple

    def __init__(self, items_field_types, **kwargs):
        self.items_fields_types = items_field_types
        super().__init__(**kwargs)

    def serialize_default(self, format, values):
        serialized_values = []
        for value, field in zip(values, self.items_fields_types):
            serialized_values.append(field.serialize(format, value))
        return serialized_values

    def deserialize_default(self, format, values):
        serialized_values = []
        for value, field in zip(values, self.items_fields_types):
            serialized_values.append(field.deserialize(format, value))
        return tuple(serialized_values)

    def _validate(self, value):
        super()._validate(value)
        if value is None:
            return  # allow_none validated in base class
        for value, field in zip(value, self.items_fields_types):
            field.validate(value)


class ReferenceField(Field):

    # TODO: chequar que las refencias tengan definida al menos una clave primaria

    def __init__(self, ref_class, **kwargs):
        self._ref_class = ref_class
        self.USE_THIN_REF = os.environ.get('M9G_SERIALIZE_THIN', False) in (
            "T", "True", True, "1", "Y", "true"
        )
        super().__init__(**kwargs)

    def _get_module_and_class(self, ref_str):
        tokens = ref_str.split('.')
        class_name = tokens.pop()
        module = '.'.join([x for x in tokens])
        return module, class_name

    @property
    def ref_class(self):
        if isinstance(self._ref_class, future_str):
            from m9g import Model
            module, class_name = self._get_module_and_class(self._ref_class)
            model = Model.get_model(module, class_name)
            self._ref_class = model
        return self._ref_class

    @property
    def FIELD_TYPE(self):
        return self.ref_class

    def _is_lazy(self, ref, check_for_thin=False):
        """ Indica si el objeto está en modo lazy. Si thin != None
            me indica también si el objeo es una thinRef """
        lazy = getattr(ref, '_ref_status', None) == 'lazy'
        if not lazy:
            return False
        if check_for_thin is False:
            return lazy
        return len(ref.available_fields()) == len(ref.primary_key_fields)

    def _build_thin_ref(self, value, model, format):
        # es de la forma (<primary_key_value>, <primary_key_value>...)
        if not isinstance(value, list):
            assert len(model.primary_key_fields) == 1
            value = [value]

        pk_values = tuple(v for v in value)
        attrs = {}

        for i, serialized_pk_value in enumerate(pk_values):
            attr_key = model.primary_key_fields[i]
            attr_value = model.field_from_key(attr_key).deserialize(format, serialized_pk_value)
            attrs[attr_key] = attr_value

        return model.thinRef(**attrs)

    def _build_gross_ref(self, value, model, format):
        # es de la forma {'<pk1>': <pk_value>...}
        fields = {}
        for k, v in iteritems(value):
            fields[k] = model.field_from_key(k).deserialize(format, v)
        return model.grossRef(**fields)

    def _build_full_obj(self, value, model, format):
        fields = {}
        for k, v in iteritems(value):
            fields[k] = model.field_from_key(k).deserialize(format, v)
        return model(**fields)

    def _validate(self, ref):
        super()._validate(ref)
        if ref is None:
            return  # allow_none validated in base class
        if not ref.primary_key_fields:
            raise ValidationError("References must have primary keys declared")

    def serialize_default(self, format, ref):
        ret = self._serialize_default(format, ref)
        if self.ref_class.polymorphic_base:
            if type(ret) == dict:
                ret["__modelclass__"] = ref.modelfullname()
            else:
                ret = {
                    "__pk__": ret,
                    "__modelclass__": ref.modelfullname()
                }
        return ret

    def _serialize_default(self, format, ref):
        # TODO: hacer validacion de tipos contra el registry

        if self.USE_THIN_REF or self._is_lazy(ref, check_for_thin=True):
            # serializo solo la pk
            if len(ref.primary_key_fields) == 1:
                pk_key = ref.primary_key_fields[0]
                pk_field = ref._fields[pk_key]

                pk_value = getattr(ref, pk_key)

                return pk_field.serialize(format, pk_value)
            else:
                return tuple(
                    ref._fields[field_name].serialize(format, getattr(ref, field_name))
                    for field_name in ref.primary_key_fields
                )
        elif self._is_lazy(ref, check_for_thin=False):
            # serializo solo los campos que hay, devuelvo diccionario
            ret = {}
            for field_name, field_value in ref.available_fields().items():
                ret[field_name] = ref._fields[field_name].serialize(format, field_value)
            return ret
        else:
            # serializo objeto completo
            return ref.serialize(self._field_format_to_model_format(format))

    def deserialize_default(self, format, value):
        if isinstance(value, dict) and "__modelclass__" in value:
            modelclass = value.pop("__modelclass__")
            model = self.ref_class.get_model(*modelclass.rsplit(".", 1))
            if "__pk__" in value:
                value = value["__pk__"]
        else:
            model = self.ref_class

        if isinstance(value, dict):
            if len(value) == len(model._fields):
                return self._build_full_obj(value, model, format)
            else:
                return self._build_gross_ref(value, model, format)

        return self._build_thin_ref(value, model, format)


class DynamicField(Field):
    FIELD_TYPE = object

    INFERRED_FIELDS = [
        StringField,
        IntField,
        FloatField,
        DecimalField,
        BytesField,
        DateField,
        DateTimeField,
        DictField,
        ListField,
        TupleField,
    ]

    def _lookup_for_field_type(self, field_type_str):
        for type_ in self.INFERRED_FIELDS:
            if type_.__name__ == field_type_str:
                return type_
        return None

    def _field_from_field_value(self, field_value):
        for field_class in self.INFERRED_FIELDS:
            # no usamos isinstance porque no discrimina entre datetime y date
            if field_class.is_field_type_of(field_value):
                return field_class
        return None

    def _build_field_as_dynamic(self, field_type, value):
        if field_type == DictField:
            return DictField(DynamicField(), DynamicField())
        if field_type == ListField:
            return ListField(DynamicField())
        if field_type == TupleField:
            args = tuple([DynamicField()] * len(value))
            return TupleField(args)
        return field_type()

    def serialize_default(self, format, value):
        field_type = self._field_from_field_value(value)

        if not field_type:
            raise ValidationError(f"{value} is not a valid value for a DynamicField")

        field = self._build_field_as_dynamic(field_type, value)
        return (field.__class__.__name__, field.serialize(format, value))

    def deserialize_default(self, format, value):
        field_type_name, serialized_value = value
        field_type = self._lookup_for_field_type(field_type_name)
        field = self._build_field_as_dynamic(field_type, value)
        return field.deserialize(format, serialized_value)

    def serialize_json(self, value):
        # JSON no soporte tuplas como claves (caso diccionario)
        return str(self.serialize_default("json", value))

    def deserialize_json(self, value):
        value = eval(value)
        return self.deserialize_default("json", value)


class CompositeField(Field):

    def __init__(self, ref_class, **kwargs):
        self.ref_class = ref_class
        self.FIELD_TYPE = ref_class
        super().__init__(**kwargs)

    def serialize_default(self, format, value):
        return value.serialize(format=self._field_format_to_model_format(format))

    def deserialize_default(self, format, value):
        return self.ref_class.deserialize(value, format=self._field_format_to_model_format(format))



# ~ coding: future_fstrings ~

from __future__ import absolute_import, unicode_literals

__version__ = "0.0.5"

from future.utils import with_metaclass, iteritems
from future.builtins import filter

import sys
import json
import pickle

from .exceptions import ValidationError
from .fields import *  # noqa


def _new_class_with_kwargs(cls, **kwargs):
    obj = cls.__new__(cls)
    for attr, attr_value in iteritems(kwargs):
        setattr(obj, attr, attr_value)
    return obj


def _get_parents_fields(parents):
    fields = {}

    for parent in parents:
        if hasattr(parent, "_fields"):
            parent_fields = dict(parent._fields)
            fields.update(parent_fields)

    return fields


class MetaModel(type):
    _models_registry = {}

    def _add_to_model_registry(cls):
        model_module, model_name = cls.module_and_classname()
        cls._models_registry[(model_module, model_name)] = cls

    def __init__(cls, class_name, parents, attrs):
        m_parents = filter(lambda p: isinstance(p, MetaModel), parents)

        if len(list(m_parents)) > 1:
            raise ValidationError(f"{cls.__name__} must only have one MetaModel parent")

        fields = _get_parents_fields(parents)

        fields_declared_as_pk = []

        for attr_k, attr_v in iteritems(attrs):
            if isinstance(attr_v, Field):  # noqa: F405
                fields[attr_k] = attr_v
                delattr(cls, attr_k)
                if attr_v.pk:
                    fields_declared_as_pk.append(attr_k)
        cls._fields = fields

        declared_pk = getattr(cls, "primary_key_fields", None)

        if declared_pk and not set(declared_pk).issubset(cls._fields):
            raise ValidationError(f"{declared_pk} is not part of declared fields")

        if len(fields_declared_as_pk) > 1:
            raise ValidationError(
                f"Multiple fields declared as primary key, got: {len(fields_declared_as_pk)}"
            )

        if len(fields_declared_as_pk) and declared_pk:
            raise ValidationError("Both fields declared as primary key and defined primary_key attr")

        field_pk = None
        if len(fields_declared_as_pk) == 1:
            field_pk = fields_declared_as_pk.pop()

        if not declared_pk:
            cls.primary_key_fields = field_pk and tuple([field_pk])

        cls._add_to_model_registry()

    @classmethod
    def get_all_models(cls):
        return cls._models_registry

    @classmethod
    def get_model(cls, module, name):
        return cls._models_registry[(module, name)]


class Model(with_metaclass(MetaModel)):
    polymorphic_base = False

    def __init__(self, **kwargs):
        received_fields = set()
        for k, v in kwargs.items():
            if k not in self._fields:
                raise ValidationError(f"'{k}' is not part of declared fields")
            setattr(self, k, v)
            received_fields.add(k)

        # seteamos a default los campos que no nos definieron
        for missing_field in set(self._fields).difference(received_fields):
            default_missing_field = self._fields[missing_field].get_default()
            setattr(self, missing_field, default_missing_field)

    @classmethod
    def module_and_classname(cls):
        model_module = getattr(cls, "model_module", None)
        if model_module is None:
            class_module = sys.modules.get(cls.__module__, None)
            if class_module and hasattr(class_module, "m9g_module"):
                model_module = class_module.m9g_module
            else:
                model_module = cls.__module__
        model_name = getattr(cls, "model_name", cls.__name__)
        return model_module, model_name

    @classmethod
    def modelfullname(cls):
        model_module, model_name = cls.module_and_classname()
        return f"{model_module}.{model_name}"

    def __getattr__(self, attr_name):
        if attr_name not in self._fields.keys():
            raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{attr_name}'")
        if self._ref_status == "lazy":
            obj = self.manager.findByPrimaryKey(self.pk)
            self._full_obj = obj
            self._ref_status = "loaded"
            return getattr(obj, attr_name)
        elif self._ref_status == "loaded":
            return getattr(self._full_obj, attr_name)

    def __setattr__(self, attr_name, attr_value):
        if attr_name not in self._fields:
            return super().__setattr__(attr_name, attr_value)

        attr_value = self._fields[attr_name].adapt(attr_value)
        self._fields[attr_name].validate(attr_value)
        super().__setattr__(attr_name, attr_value)

    def _resolve_pk_values(self):
        """ Obtiene los valores de las claves primarias """
        return [getattr(self, attr_name) for attr_name in self.primary_key_fields if self.primary_key_fields]

    def _serialize(self, format):
        """ Serialización completa, con todos los atributos """

        data = {}
        format = self._model_format_to_field_format(format)
        for k, v in iteritems(self._fields):
            value = getattr(self, k)
            data[k] = v.serialize(format, value)

        if self.polymorphic_base:
            data["__modelclass__"] = self.modelfullname()

        return data

    @classmethod
    def _model_format_to_field_format(cls, format):
        if format == "jsondict":
            return "json"
        return format

    @classmethod
    def _deserialize(cls, data, format):
        modelclass = data.pop("__modelclass__", None)
        if modelclass:
            model_module, model_class = modelclass.rsplit(".", 1)
            subclass = MetaModel.get_model(model_module, model_class)
            return subclass._deserialize(data, format)

        return _new_class_with_kwargs(cls, **cls._deserialize_fields(data, format))

    def _in_place_deserialize(self, data, format):
        new_fields = self._deserialize_fields(data, format)
        for k, v in new_fields.items():
            setattr(self, k, v)

    @classmethod
    def _deserialize_fields(cls, data, format):
        """Full deserialization - expects all the fields"""
        kwattrs = {}
        format = cls._model_format_to_field_format(format)
        for attr_name, serialized_attr_value in iteritems(data):
            if attr_name not in cls._fields.keys():
                # TODO: ver qué pasa si nos "sobran" atributos en data - WARNING o ERROR?
                pass

            associated_field = cls._fields[attr_name]
            attr_value = associated_field.deserialize(format, serialized_attr_value)
            kwattrs[attr_name] = attr_value

        return kwattrs

    @property
    def pk(self):
        pk_values = self._resolve_pk_values()
        if len(pk_values) == 1:
            return pk_values[0]
        return tuple(pk_values)

    def available_fields(self):
        """ Returns dictionary with <field_type>: <field_value> for fields available in the
            gross or thin ref """
        return dict((field_name, field_value) for field_name, field_value in self.__dict__.items()
                    if field_name in self._fields)

    @classmethod
    def field_from_key(cls, field_key):
        return cls._fields[field_key]

    @classmethod
    def get_manager(cls):
        return cls.manager

    def serialize(self, format="json"):
        serialize_method = getattr(self, f"serialize_{format}")
        return serialize_method()

    def serialize_json(self):
        """Serializes as JSON string"""
        return json.dumps(self.serialize_jsondict())

    def serialize_jsondict(self):
        """Serializes as JSON-serializable python dictionary"""
        return self._serialize("jsondict")

    def serialize_pickle(self):
        serialized_data = self._serialize("pickle")
        return pickle.dumps(serialized_data)

    def serialize_pydict(self):
        return self._serialize("pydict")

    @classmethod
    def deserialize(cls, str_obj, format="json"):
        deserialize_method = getattr(cls, f"deserialize_{format}")
        return deserialize_method(str_obj)

    def in_place_deserialize(self, str_obj, format="json"):
        deserialize_method = getattr(self, f"in_place_deserialize_{format}")
        return deserialize_method(str_obj)

    def in_place_deserialize_json(self, str_obj):
        data = json.loads(str_obj)
        return self.in_place_deserialize_jsondict(data)

    def in_place_deserialize_jsondict(self, obj):
        return self._in_place_deserialize(obj, "jsondict")

    def in_place_deserialize_pickle(self, str_obj):
        data = pickle.loads(str_obj)
        return self._in_place_deserialize(data, "pickle")

    def in_place_deserialize_pydict(self, pydict):
        return self._in_place_deserialize(pydict, "pydict")

    @classmethod
    def deserialize_json(cls, str_obj):
        data = json.loads(str_obj)
        return cls.deserialize_jsondict(data)

    @classmethod
    def deserialize_jsondict(cls, obj):
        return cls._deserialize(obj, "jsondict")

    @classmethod
    def deserialize_pickle(cls, str_obj):
        data = pickle.loads(str_obj)
        return cls._deserialize(data, "pickle")

    @classmethod
    def deserialize_pydict(cls, pydict):
        return cls._deserialize(pydict, "pydict")

    @classmethod
    def fromdb(cls, **kwargs):
        if set(kwargs) != set(cls._fields):
            raise ValidationError("Faltan")

        obj = _new_class_with_kwargs(cls, **kwargs)
        # TODO: marca de readonly?
        return obj

    @classmethod
    def thinRef(cls, **kwargs):
        if set(kwargs) != set(cls.primary_key_fields):
            raise ValidationError(f"{cls.__name__}.thinRef must receive only primary key fields")

        obj = _new_class_with_kwargs(cls, **kwargs)
        obj._ref_status = "lazy"
        return obj

    @classmethod
    def grossRef(cls, **kwargs):
        received_fields = set(kwargs)
        if not received_fields.issuperset(set(cls.primary_key_fields)):
            raise ValidationError(f"{cls.__name__}.grossRef must receive at least primary keys")
        if not set(cls._fields).issuperset(received_fields):
            unknown_fields = received_fields - set(cls._fields)
            raise ValidationError(f"{cls.__name__}.grossRef received unknown_fields: {unknown_fields}")

        obj = _new_class_with_kwargs(cls, **kwargs)
        obj._ref_status = "lazy"
        return obj

    def is_lazy(self):
        return hasattr(self, "_ref_status") and self._ref_status == "lazy"
